import time
from datetime import datetime, timedelta


class TimeSync:
    def __init__(self) -> None:
        self.name = "timeSync"
        self.server_timestamp = None
        self.expiration_time = 1

    @property
    def server_datetime(self):
        while self.server_timestamp is None:
            time.sleep(0.2)
        return datetime.fromtimestamp(self.server_timestamp)

    @property
    def expiration_datetime(self):
        return self.server_datetime + timedelta(minutes=self.expiration_time)

    @property
    def expiration_timestamp(self):
        return time.mktime(self.expiration_datetime.timetuple())



class Profile:
    def __init__(self):
        self.name = "profile"
        self.id = None
        self.skey = None
        self.balance = None
        self.balance_id = None
        self.balances: list[dict] = None
        self.msg = None
        self.currency = None
        self.minimum_amount = 1
        self.balance_type = None
        self.currency_char = None
        self.time_zone = -3



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def profile(api: "IQOptionAPI", message: dict):
    if message.get("name") != "profile":
        return

    api.profile.msg = message.get("msg")
    if api.profile.msg is None or api.profile.msg is False:
        return

    msg: dict = message.get("msg", {})

    api.profile.id = msg.get("user_id")
    api.profile.balance_id = msg.get("balance_id")
    api.profile.balance_type = msg.get("balance_type")
    api.profile.balances = msg.get("balances")



import json
import logging
import ssl

from websockets.asyncio.client import connect

from iqoption_async.ws.received.balances import balances
from iqoption_async.ws.received.candles import candles
from iqoption_async.ws.received.initialization_data import initialization_data
from iqoption_async.ws.received.option import option
from iqoption_async.ws.received.position_changed import position_changed
from iqoption_async.ws.received.profile import profile
from iqoption_async.ws.received.result import result
from iqoption_async.ws.received.time_sync import time_sync


class WebSocketClient:
    def __init__(self, api) -> None:
        self.api = api
        self.uri = self.api.wss_url
        self.logger = logging.getLogger(__name__)
        self.ssl_context = ssl.create_default_context()
        self.ssl_context.check_hostname = False
        self.ssl_context.verify_mode = ssl.CERT_NONE
        self.websocket = None
        self.websocket_error_reason = None
        self.websocket_has_error = False
        self.websocket_connected = None
        self.user_agent_header = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36"  # noqa: E501

    async def check_is_connected(self):
        try:
            await self.websocket.ping()
            return True
        except Exception:
            return False

    async def connect_and_run(self):
        try:
            async with connect(
                self.uri, ssl=self.ssl_context, max_size=None, user_agent_header=self.user_agent_header
            ) as websocket:
                self.websocket = websocket
                self.logger.debug("Websocket client connected.")
                self.websocket_connected = True
                await self.receive()
        except Exception as e:
            self.logger.debug(f"Websocket error {e}")
            self.websocket_error_reason = str(e)
            self.websocket_has_error = True
        finally:
            self.logger.debug("Websocket connection closed.")
            self.websocket_connected = False

    async def receive(self):
        try:
            async for message in self.websocket:
                await self.on_message(message)
        except Exception as e:
            self.logger.error(f"Error during receiving messages: {e}")
            self.websocket_error_reason = str(e)
            self.websocket_has_error = True

    async def on_message(self, message):
        message = json.loads(str(message))

        result(self.api, message)
        candles(self.api, message)
        profile(self.api, message)
        time_sync(self.api, message)
        position_changed(self.api, message)
        balances(self.api, message)
        initialization_data(self.api, message)
        option(self.api, message)



import asyncio
import logging
import sys
import time
from collections import defaultdict
from datetime import datetime, timedelta
from random import randint

import pytz

from iqoption_async.api import IQOptionAPI
from iqoption_async.constants import ACTIVES
from iqoption_async.version_control import API_VERSION


def nested_dict(n, type):
    if n == 1:
        return defaultdict(type)
    else:
        return defaultdict(lambda: nested_dict(n - 1, type))


class IQOptionClient:
    __version__ = API_VERSION
    api: IQOptionAPI

    OPEN_TIME = nested_dict(3, dict)
    PROFITS = nested_dict(3, dict)

    def __init__(self, email: str, password: str):
        self.email = email
        self.password = password
        self.logger = logging.getLogger(__name__)

        agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36"  # noqa: E501
        self.session_header = {"User-Agent": agent}
        self.session_cookie = {}

    async def close(self):
        await self.api.close()

    async def connect(self):
        self.api = IQOptionAPI("iqoption.com", self.email, self.password)
        self.api.set_session(self.session_cookie, self.session_header)

        check, reason = await self.api.connect()
        if not check:
            return False, reason

        while self.api.profile.balance_id is None:
            await asyncio.sleep(0.1)

        tasks = []
        tasks.append(self.position_change_all("subscribeMessage", self.api.profile.balance_id))
        tasks.append(self.order_changed_all("subscribeMessage"))
        tasks.append(self.api.set_options(1, True))
        tasks.append(self.get_all_open_time())
        await asyncio.gather(*tasks)
        return True, None

    async def position_change_all(self, main_name: str, balance_id: str):
        instrument_type = ["cfd", "forex", "crypto", "digital-option", "turbo-option", "binary-option"]
        tasks = []
        for ins in instrument_type:
            tasks.append(
                self.api.portfolio(main_name, "portfolio.position-changed", ins, balance_id, self.api.profile.id)
            )
        await asyncio.gather(*tasks)

    async def order_changed_all(self, main_name: str):
        instrument_type = [
            "cfd",
            "forex",
            "crypto",
            "digital-option",
            "turbo-option",
            "binary-option",
            "turbo",
            "binary",
        ]
        tasks = []
        for ins in instrument_type:
            tasks.append(self.api.portfolio(main_name, "portfolio.order-changed", ins, user_id=self.api.profile.id))
        await asyncio.gather(*tasks)

    async def check_connect(self):
        return await self.api.websocket_client.check_is_connected()

    async def get_profile(self):
        while self.api.profile.msg is None:
            await asyncio.sleep(0.1)
        return self.api.profile

    async def change_balance(self, balance_mode: str):
        real_id = None
        practice_id = None
        tournament_id = None

        profile = await self.get_profile()
        for balance in profile.balances:
            balance_type = balance.get("type")
            balance_id = balance.get("id")
            if balance_type == 1:
                real_id = balance_id
            if balance_type == 4:
                practice_id = balance_id
            if balance_type == 2:
                tournament_id = balance_id

        if balance_mode == "REAL":
            await self.__set_balance_mode(real_id)
        elif balance_mode == "PRACTICE":
            await self.__set_balance_mode(practice_id)
        elif balance_mode == "TOURNAMENT":
            await self.__set_balance_mode(tournament_id)
        else:
            self.logger.error(f"ERROR {balance_mode} does not exists")
            await self.close()
            sys.exit(1)

    async def __set_balance_mode(self, balance_id: int):
        if self.api.profile.balance_id == balance_id:
            return
        if self.api.profile.balance_id is not None:
            await self.position_change_all("unsubscribeMessage", self.api.profile.balance_id)
        self.api.profile.balance_id = balance_id
        await self.position_change_all("subscribeMessage", balance_id)

    async def get_balance(self):
        if self.api.profile.balance_id is None:
            self.logger.error("ERROR balance account not configured")
            return None
        balances_raw = await self.get_balances()
        for balance in balances_raw.get("msg"):
            if balance["id"] == self.api.profile.balance_id:
                return balance["amount"]
        return None

    async def get_balances(self):
        self.api.balances_raw = None
        await self.api.get_balances()
        while self.api.balances_raw is None:
            await asyncio.sleep(0.1)
        return self.api.balances_raw

    async def get_all_open_time(self):
        self.OPEN_TIME = nested_dict(3, dict)
        tasks = [self.__get_binary_open()]
        await asyncio.gather(*tasks)
        return self.OPEN_TIME

    async def get_all_init(self):
        self.api.api_option_init_all_result = None
        if not (await self.check_connect()):
            await self.connect()

        await self.api.get_api_option_init_all()
        start_at = time.time()
        retries = 0
        max_retries = 3
        while True:
            while self.api.api_option_init_all_result is None:
                if time.time() - start_at >= 30:
                    self.logger.error("WARN get_api_option_init_all late 30 sec")
                    break
                await asyncio.sleep(0.1)
            if self.api.api_option_init_all_result is not None:
                break
            retries += 1
            if retries >= max_retries:
                return None
            await asyncio.sleep(0.1)

        return self.api.api_option_init_all_result

    async def __get_binary_open(self):
        binary_data = await self.get_all_init()
        binary_list = ["binary", "turbo"]
        if binary_data is None:
            return
        for option in binary_list:
            if option not in binary_data:
                continue

            actives: dict[int, dict] = binary_data[option]["actives"]
            for i, details in actives.items():
                name_part = str(details["name"].split(".")[1])
                ACTIVES[name_part] = int(i)

                is_open = details.get("enabled", False) and not details.get("is_suspended", True)
                self.OPEN_TIME[option][name_part]["open"] = is_open
                self.OPEN_TIME[option][name_part]["schedules"] = details.get("schedule", [])

                option_profit = details.get("option", {"profit": {"commission": 100}})["profit"]
                commission = option_profit["commission"]
                profit = (100.0 - commission) / 100.0
                self.PROFITS[option][name_part]["profit"] = profit

    async def __check_order(self, order_id: int, expiration: int):
        expiration = datetime.now(pytz.utc) + timedelta(minutes=expiration, seconds=10)
        while True:
            now = datetime.now(pytz.utc)
            if now >= expiration:
                self.logger.warning(f"ERROR order not found after {expiration} mins")
                return False, None, None

            order = self.api.order_async.get(order_id)
            if order is None:
                await asyncio.sleep(0.1)
                continue
            detail: dict = order.get("position-changed", {}).get("msg")
            if detail is None:
                await asyncio.sleep(0.1)
                continue
            status = detail.get("status")
            close_reason = detail.get("close_reason")
            pnl = detail.get("pnl")
            if status != "closed" or close_reason is None or pnl is None:
                await asyncio.sleep(0.1)
                continue
            return True, close_reason, pnl

    async def check_order(self, order_id: int, expiration: int = 1, async_fnc=None):
        order_id = int(order_id)

        async def call_fnc(_fnc, _order_id: int, _expiration: int):
            result = await self.__check_order(_order_id, _expiration)
            await _fnc(result)

        if async_fnc is not None:
            asyncio.create_task(call_fnc(async_fnc, order_id, expiration))
            return

        return await self.__check_order(order_id, expiration)

    async def get_candles(self, symbol: str, interval_min: int, limit: int, end_time: float):
        if symbol not in ACTIVES:
            self.logger.warning(f"Asset {symbol} not found")
            return None

        request_id = self.api.generate_request_id()
        self.api.candles.add_request_id(request_id)
        interval_min *= 60
        await self.api.get_candles(ACTIVES[symbol], interval_min, limit, end_time, request_id)
        candles_data = self.api.candles.get_candles_data(request_id)
        while candles_data is None:
            await asyncio.sleep(0.1)
            candles_data = self.api.candles.get_candles_data(request_id)
            if not (await self.check_connect()):
                self.logger.error("ERROR get_candles need reconnect")
                break
        return candles_data

    async def get_profit(self, symbol: str, option_type: str):
        par = self.PROFITS.get(option_type, {}).get(symbol)
        if par is None:
            return None
        return par["profit"]

    async def buy(self, amount: float, symbol: str, action: str, expiration: int):
        if self.api.profile.balance_id is None:
            self.logger.error("ERROR balance account not configured")
            return False, None
        if symbol not in ACTIVES:
            self.logger.warning(f"Asset {symbol} not found")
            return False, None
        request_id = str(randint(0, 10000))
        active_id = ACTIVES[symbol]
        action = action.lower()
        self.api.result[request_id] = {"result": None, "id": None}

        await self.api.buy(float(amount), active_id, action, expiration, request_id)
        start_at = datetime.now(pytz.utc) + timedelta(seconds=5)
        order_id = None
        order_result = None
        while True:
            now = datetime.now(pytz.utc)
            if now >= start_at:
                self.logger.error("WARNINING buy late 5 sec")
                print(self.api.result)
                return False, None, None

            order = self.api.result[request_id]
            message = order.get("message")
            if message is not None:
                self.api.result.pop(request_id)
                return False, message, None

            order_id = order.get("id")
            order_expiration = order.get("exp")
            order_result = order.get("result")
            if order_id is not None and order_result is not None:
                self.api.result.pop(request_id)
                return order_result, order_id, order_expiration
            await asyncio.sleep(0.1)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def initialization_data(api: "IQOptionAPI", message: dict):
    if message.get("name") == "initialization-data":
        api.api_option_init_all_result = message.get("msg")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def balances(api: "IQOptionAPI", message: dict):
    if message.get("name") == "balances":
        api.balances_raw = message



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


class Resource:
    url = ""

    def __init__(self, api: "IQOptionAPI"):
        self.api = api

    async def send_http_request(self, method: str, data: dict = None, params: dict = None, headers: dict = None):
        return await self.api.send_http_request(self, method, data, params, headers)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def option(api: "IQOptionAPI", message: dict):
    if message.get("name") == "option":
        request_id = message.get("request_id")
        if request_id in api.result:
            msg: dict = message.get("msg", {})
            for key, val in msg.items():
                if key in ["id", "message", "exp"]:
                    api.result[request_id][key] = val



import asyncio
import json
import logging
import time
from collections import defaultdict

import aiohttp

from iqoption_async.expiration import get_expiration_time
from iqoption_async.http.resource import Resource
from iqoption_async.ws.client import WebSocketClient
from iqoption_async.ws.objects.candles import Candles
from iqoption_async.ws.objects.profile import Profile
from iqoption_async.ws.objects.time_sync import TimeSync


def nested_dict(n, type):
    if n == 1:
        return defaultdict(type)
    else:
        return defaultdict(lambda: nested_dict(n - 1, type))


class IQOptionAPI:
    profile = Profile()
    timesync = TimeSync()
    candles = Candles()
    order_async = nested_dict(2, dict)
    position_changed = None
    balances_raw: dict = None
    api_option_init_all_result: list = []
    result: dict[str, dict] = {}

    def __init__(self, host: str, username: str, password: str):
        self.https_url = f"https://{host}/api"
        self.wss_url = f"wss://{host}/echo/websocket"
        self.username = username
        self.password = password
        self.session = aiohttp.ClientSession(trust_env=False)
        self.ssl_verify = False
        self.logger = logging.getLogger(__name__)
        self.lock = asyncio.Lock()

    async def close(self):
        await self.websocket_client.websocket.close()

    def set_session(self, cookies: dict, headers: dict):
        self.session.headers.update(headers)
        jar = self.session.cookie_jar
        jar.clear()
        for name, value in cookies.items():
            jar.update_cookies({name: value})

    async def connect(self):
        try:
            await self.close()
        except Exception:
            pass

        check_websocket, websocket_reason = await self.start_websocket()
        if not check_websocket:
            return check_websocket, websocket_reason

        response = await self.get_ssid()
        if isinstance(response, Exception):
            await self.close()
            return False, str(response)

        ssid = response.cookies.get("ssid", None)
        if ssid is None:
            return False, "SSID cookie not found"

        ssid = ssid.value
        await self.send_ssid(ssid)
        self.session.cookie_jar.update_cookies({"ssid": ssid})

        self.timesync.server_timestamp = None
        while self.timesync.server_timestamp is None:
            await asyncio.sleep(0.1)
        return True, None

    async def start_websocket(self):
        self.websocket_client = WebSocketClient(self)
        self.websocket_client_task = asyncio.create_task(self.websocket_client.connect_and_run())

        while True:
            await asyncio.sleep(1)
            if self.websocket_client.websocket_has_error:
                return False, self.websocket_client.websocket_error_reason
            if self.websocket_client.websocket_connected is True:
                return True, None
            if self.websocket_client.websocket_connected is False:
                return False, "Websocket connection closed."

    async def get_ssid(self):
        url = "https://auth.iqoption.com/api/v2/login"
        data = {"identifier": self.username, "password": self.password}
        try:
            return await self.send_http_request_v2(url, "POST", data=data)
        except Exception as e:
            self.logger.error(e)
            return e

    async def send_ssid(self, ssid):
        await self.send_websocket_request("ssid", ssid)
        while self.profile.msg is None:
            await asyncio.sleep(0.1)
        return self.profile.msg is not False

    def prepare_http_url(self, resource: Resource):
        return "/".join((self.https_url, resource.url))

    async def send_http_request(
        self, resource: Resource, method: str, data: dict = None, params: dict = None, headers: dict = None
    ):
        url = self.prepare_http_url(resource)
        self.logger.debug(
            f"{method}: {url} headers {str(self.session.headers)} cookies: {str(self.session.cookie_jar)}"
        )

        async with self.session.request(method=method, url=url, data=data, params=params, headers=headers) as response:
            self.logger.debug(response)
            self.logger.debug(await response.text())
            self.logger.debug(response.headers)
            self.logger.debug(response.cookies)
            response.raise_for_status()
            return response

    async def send_http_request_v2(
        self, url: str, method: str, data: dict = None, params: dict = None, headers: dict = None
    ):
        self.logger.debug(
            f"{method}: {url} headers {str(self.session.headers)} cookies: {str(self.session.cookie_jar)}"
        )

        async with self.session.request(method=method, url=url, data=data, params=params, headers=headers) as response:
            self.logger.debug(response)
            self.logger.debug(await response.text())
            self.logger.debug(response.headers)
            self.logger.debug(response.cookies)
            return response

    def generate_request_id(self):
        return str(int(str(time.time()).split(".")[1]))

    async def send_websocket_request(self, name: str, msg: dict, request_id: str = ""):
        if request_id == "":
            request_id = self.generate_request_id()

        request_id = str(request_id)
        data = json.dumps(dict(name=name, msg=msg, request_id=request_id))
        self.logger.debug(data)

        async with self.lock:
            await self.websocket_client.websocket.send(data)

    async def portfolio(
        self,
        main_name: str,
        name: str,
        instrument_type: str,
        balance_id: int = None,
        user_id: int = None,
        limit: int = 1,
        offset: int = 0,
        request_id: str = "",
    ):
        msg: dict = None
        if name == "portfolio.order-changed":
            msg = {
                "name": name,
                "version": "2.0",
                "params": {"routingFilters": {"instrument_type": instrument_type, "user_id": user_id}},
            }
        elif name == "portfolio.get-positions":
            msg = {
                "name": name,
                "version": "4.0",
                "body": {
                    "instrument_type": instrument_type,
                    "limit": limit,
                    "offset": offset,
                    "user_balance_id": balance_id,
                },
            }
        elif name == "portfolio.position-changed":
            msg = {
                "name": name,
                "version": "3.0",
                "params": {
                    "routingFilters": {
                        "instrument_type": instrument_type,
                        "user_balance_id": balance_id,
                        "user_id": user_id,
                    }
                },
            }

        if msg is None:
            return
        await self.send_websocket_request(main_name, msg, request_id)

    async def set_options(self, request_id: str, send_results):
        msg = {"sendResults": send_results}
        await self.send_websocket_request("setOptions", msg, request_id)

    async def get_balances(self):
        msg = {"name": "get-balances", "version": "1.0"}
        await self.send_websocket_request("sendMessage", msg)

    async def get_api_option_init_all(self):
        msg = {"name": "get-initialization-data", "version": "4.0", "body": {}}
        await self.send_websocket_request("sendMessage", msg)

    async def get_candles(self, active_id: int, interval: int, count: int, end_time: float, request_id: str):
        msg = {
            "name": "get-candles",
            "version": "2.0",
            "body": {
                "active_id": active_id,
                "split_normalization": True,
                "size": interval,
                "count": count,
                "to": int(end_time),
            },
        }
        await self.send_websocket_request("sendMessage", msg, request_id)

    async def buy(self, amount: int, active_id: int, direction: str, expiration: int, request_id: str):
        exp, idx = get_expiration_time(int(self.timesync.server_timestamp), expiration)
        option = None
        if idx < 5:
            option = 3
        else:
            option = 1

        msg = {
            "name": "binary-options.open-option",
            "version": "1.0",
            "body": {
                "price": amount,
                "active_id": active_id,
                "expired": int(exp),
                "direction": direction,
                "option_type_id": option,
                "user_balance_id": int(self.profile.balance_id),
            },
        }
        await self.send_websocket_request("sendMessage", msg, request_id)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def time_sync(api: "IQOptionAPI", message: dict):
    if message.get("name") == "timeSync":
        server_timestamp = message.get("msg")
        if server_timestamp is not None:
            api.timesync.server_timestamp = server_timestamp / 1000



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def result(api: "IQOptionAPI", message: dict):
    if message.get("name") == "result":
        request_id = message.get("request_id")
        if request_id in api.result:
            api.result[request_id]["result"] = message.get("msg", {"success": None})["success"]



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def candles(api: "IQOptionAPI", message: dict):
    if message.get("name") == "candles":
        request_id = message.get("request_id")
        if api.candles.has_request_id(request_id):
            api.candles.set_candles_data(request_id, message.get("msg")["candles"])



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


def position_changed(api: "IQOptionAPI", message: dict):
    name = message.get("name")
    if name != "position-changed":
        return

    msg: dict = message.get("msg", {})
    microservice_name: str = message.get("microserviceName")
    source = msg.get("source")

    if microservice_name == "portfolio":
        if source in ["digital-options", "trading"]:
            order_id = int(msg.get("raw_event")["order_ids"][0])
            api.order_async[order_id][name] = message
            return
        elif source == "binary-options":
            order_id = int(msg.get("external_id"))
            api.order_async[order_id][name] = message
            return

    api.position_changed = message



from setuptools import setup, find_packages

setup(
    name="iqoption-async",
    version="0.0.1",
    description="An updated Python library for interacting with IQ Option using asyncio and modern WebSockets.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Victor Rivera",
    author_email="victor.riverac92@gmail.com",
    url="https://github.com/s00rk/iqoption_async",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.11",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="iqoption trading api asyncio websockets"
)



class Candles:
    def __init__(self):
        self.name = "candles"
        self.__candles_data = {}

    def get_candles_data(self, request_id: str):
        if self.__candles_data.get(request_id) is None:
            return None
        return self.__candles_data.pop(request_id, None)

    def set_candles_data(self, request_id: str, candles_data):
        self.__candles_data[request_id] = candles_data

    def add_request_id(self, request_id: str):
        if not self.has_request_id(request_id):
            self.__candles_data[request_id] = None

    def has_request_id(self, request_id: str):
        return request_id in self.__candles_data



import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iqoption_async.api import IQOptionAPI


class Base:
    def __init__(self, api: "IQOptionAPI"):
        self.api = api

    def send_websocket_request(self, name: str, msg: dict, request_id: str = ""):
        if request_id == "":
            request_id = str(int(str(time.time()).split(".")[1]))
        return self.api.send_websocket_request(name, msg, request_id)



import time
from datetime import datetime, timedelta


def date_to_timestamp(dt):
    return time.mktime(dt.timetuple())


def get_expiration_time(timestamp, duration):
    #
    now_date = datetime.fromtimestamp(timestamp)
    exp_date = now_date.replace(second=0, microsecond=0)
    if (int(date_to_timestamp(exp_date + timedelta(minutes=1))) - timestamp) > 30:
        exp_date = exp_date + timedelta(minutes=1)

    else:
        exp_date = exp_date + timedelta(minutes=2)
    exp = []
    for _ in range(5):
        exp.append(date_to_timestamp(exp_date))
        exp_date = exp_date + timedelta(minutes=1)

    idx = 50
    index = 0
    now_date = datetime.fromtimestamp(timestamp)
    exp_date = now_date.replace(second=0, microsecond=0)
    while index < idx:
        if int(exp_date.strftime("%M")) % 15 == 0 and (int(date_to_timestamp(exp_date)) - int(timestamp)) > 60 * 5:
            exp.append(date_to_timestamp(exp_date))
            index = index + 1
        exp_date = exp_date + timedelta(minutes=1)

    remaning = []

    for t in exp:
        remaning.append(int(t) - int(time.time()))

    close = [abs(x - 60 * duration) for x in remaning]

    return int(exp[close.index(min(close))]), int(close.index(min(close)))
